@@ -8,23 +8,17 @@
  * provided with absolutely no warranty. See the COPYING and CREDITS
  * files for more information.
  */
-#include "sbcl.h"
+#include "genesis/sbcl.h"
 
 #ifdef LISP_FEATURE_SB_SAFEPOINT /* entire file */
 #include <stdlib.h>
 #include <stdio.h>
 #include <string.h>
 #include <stdarg.h>
-#ifndef LISP_FEATURE_WIN32
-#include <sched.h>
-#endif
 #include <signal.h>
 #include <stddef.h>
 #include <errno.h>
 #include <sys/types.h>
-#ifndef LISP_FEATURE_WIN32
-#include <sys/wait.h>
-#endif
 #include "runtime.h"
 #include "validate.h"
 #include "thread.h"
@@ -32,9 +26,8 @@
 #include "target-arch-os.h"
 #include "os.h"
 #include "globals.h"
-#include "dynbind.h"
 #include "genesis/cons.h"
-#include "genesis/fdefn.h"
+#include "genesis/symbol.h"
 #include "interr.h"
 #include "genesis/sap.h"
 #include "gc.h"
@@ -109,6 +102,8 @@ const char* gc_phase_names[GC_NPHASES] = {
     ignore_value(mutex_acquire(&all_threads_lock)); \
     RUN_BODY_ONCE(all_threads_lock, ignore_value(mutex_release(&all_threads_lock)))
 
+extern void map_gc_page();
+extern void unmap_gc_page();
 #if !defined(LISP_FEATURE_WIN32)
 /* win32-os.c covers these, but there is no unixlike-os.c, so the normal
  * definition goes here.  Fixme: (Why) don't these work for Windows?
@@ -170,7 +165,8 @@ void safepoint_init()
     for (i=GC_NONE; i<GC_NPHASES; ++i)
         InitializeConditionVariable(&gc_state.phase_cond[i]);
     InitializeCriticalSection(&gc_state.lock);
-#else
+#elif !defined LISP_FEATURE_X86_64 && !defined LISP_FEATURE_RELOCATABLE_STATIC_SPACE
+    // 64-bit already allocated a trap page when the GC card mark table was made
     os_alloc_gc_space(0, NOT_MOVABLE, GC_SAFEPOINT_PAGE_ADDR, BACKEND_PAGE_BYTES);
 #endif
     gc_state.phase = GC_NONE;
@@ -180,7 +176,7 @@ void
 gc_state_lock()
 {
     odxprint(safepoints,"GC state to be locked");
-    int result = mutex_acquire(&gc_state.lock);
+    int __attribute__((unused)) result = mutex_acquire(&gc_state.lock);
     gc_assert(result);
     if (gc_state.master) {
         fprintf(stderr,"GC state lock glitch [%p] in thread %p phase %d (%s)\n",
@@ -201,7 +197,7 @@ gc_state_unlock()
              gc_state.phase, gc_phase_names[gc_state.phase]);
     gc_assert(get_sb_vm_thread()==gc_state.master);
     gc_state.master = NULL;
-    int result = mutex_release(&gc_state.lock);
+    int __attribute__((unused)) result = mutex_release(&gc_state.lock);
     gc_assert(result);
     odxprint(safepoints,"%s","GC state unlocked");
 }
@@ -291,7 +287,7 @@ thread_blocks_gc(struct thread *thread)
 static inline bool
 set_thread_csp_access(struct thread* th, bool writable)
 {
-    os_protect((char*)th - (THREAD_HEADER_SLOTS*N_WORD_BYTES) - THREAD_CSP_PAGE_SIZE,
+    os_protect((char*)th - THREAD_CSP_PAGE_SIZE,
                THREAD_CSP_PAGE_SIZE,
                writable? (OS_VM_PROT_READ|OS_VM_PROT_WRITE)
                : (OS_VM_PROT_READ));
@@ -464,7 +460,6 @@ thread_register_gc_trigger()
     }
 }
 
-#ifdef LISP_FEATURE_SB_SAFEPOINT
 static inline int
 thread_may_thrupt(os_context_t *ctx)
 {
@@ -563,7 +558,6 @@ check_pending_thruptions(os_context_t *ctx)
 
     return 1;
 }
-#endif
 
 int
 on_stack_p(struct thread *th, void *esp)
@@ -573,26 +567,12 @@ on_stack_p(struct thread *th, void *esp)
         < (void *)th->control_stack_end;
 }
 
-#ifndef LISP_FEATURE_WIN32
-/* (Technically, we still allocate an altstack even on Windows.  Since
- * Windows has a contiguous stack with an automatic guard page of
- * user-configurable size instead of an alternative stack though, the
- * SBCL-allocated altstack doesn't actually apply and won't be used.) */
-int
-on_altstack_p(struct thread *th, void *esp)
-{
-    void *start = (char *)th+dynamic_values_bytes;
-    void *end = (char *)start + 32*SIGSTKSZ;
-    return start <= esp && esp < end;
-}
-#endif
-
-void
-assert_on_stack(struct thread *th, void *esp)
+void assert_on_stack(struct thread *th, void *esp)
 {
     if (on_stack_p(th, esp))
         return;
-#ifndef LISP_FEATURE_WIN32
+#ifdef LISP_FEATURE_UNIX
+    extern int on_altstack_p(struct thread *th, void *esp);
     if (on_altstack_p(th, esp))
         lose("thread %p: esp on altstack: %p", th, esp);
 #endif
@@ -611,7 +591,7 @@ static bool can_invoke_post_gc(struct thread* th)
 }
 
 // returns 0 if skipped, 1 otherwise
-int check_pending_gc(__attribute((unused)) os_context_t *ctx)
+int check_pending_gc()
 {
     odxprint(misc, "check_pending_gc");
     struct thread * self = get_sb_vm_thread();
@@ -719,10 +699,8 @@ void thread_in_lisp_raised(os_context_t *ctxptr)
     /* If we still need to GC, and it's not inhibited, call into
      * SUB-GC.  Phase is either GC_QUIET or GC_NONE. */
     if (check_gc_and_thruptions) {
-        check_pending_gc(ctxptr);
-#ifdef LISP_FEATURE_SB_SAFEPOINT
+        check_pending_gc();
         while(check_pending_thruptions(ctxptr));
-#endif
     }
 }
 
@@ -778,11 +756,9 @@ void thread_in_safety_transition(os_context_t *ctxptr)
             }
         }
     }
-#ifdef LISP_FEATURE_SB_SAFEPOINT
     if (was_in_alien) {
         while(check_pending_thruptions(ctxptr));
     }
-#endif
 }
 
 #ifdef LISP_FEATURE_WIN32
@@ -805,10 +781,8 @@ void thread_interrupted(os_context_t *ctxptr)
             thread_in_lisp_raised(ctxptr);
         }
     }
-    check_pending_gc(ctxptr);
-#ifdef LISP_FEATURE_SB_SAFEPOINT
+    check_pending_gc();
     while(check_pending_thruptions(ctxptr));
-#endif
 }
 #endif
 
@@ -878,7 +852,6 @@ void gc_start_the_world()
 }
 
 
-#ifdef LISP_FEATURE_SB_SAFEPOINT
 /* wake_thread(thread) -- ensure a thruption delivery to
  * `thread'. */
 
@@ -891,7 +864,7 @@ wake_thread_io(struct thread * thread)
     win32_maybe_interrupt_io(thread);
 }
 
-void wake_thread_impl(struct thread_instance *lispthread)
+static void wake_thread_impl(struct thread_instance *lispthread)
 {
     struct thread* thread = (void*)lispthread->uw_primitive_thread;
     wake_thread_io(thread);
@@ -920,7 +893,7 @@ void wake_thread_impl(struct thread_instance *lispthread)
     return;
 }
 # else
-void wake_thread_impl(struct thread_instance *lispthread)
+static void wake_thread_impl(struct thread_instance *lispthread)
 {
     struct thread *thread = (void*)lispthread->uw_primitive_thread;
     struct thread *self = get_sb_vm_thread();
@@ -979,7 +952,63 @@ void wake_thread_impl(struct thread_instance *lispthread)
     thread_sigmask(SIG_SETMASK, &oldset, 0);
 }
 #endif /* !LISP_FEATURE_WIN32 */
-#endif /* LISP_FEATURE_SB_SAFEPOINT */
+
+/* If the thread id given does not belong to a running thread (it has
+ * exited or never even existed) pthread_kill _may_ fail with ESRCH,
+ * but it is also allowed to just segfault, see
+ * <http://udrepper.livejournal.com/16844.html>.
+ *
+ * Relying on thread ids can easily backfire since ids are recycled
+ * (NPTL recycles them extremely fast) so a signal can be sent to
+ * another process if the one it was sent to exited.
+ *
+ * For these reasons, we must make sure that the thread is still alive
+ * when the pthread_kill is called and return if the thread is
+ * exiting.
+ *
+ * Note (DFL, 2011-06-22): At the time of writing, this function is only
+ * used for INTERRUPT-THREAD, hence the wake_thread special-case for
+ * Windows is OK. */
+void wake_thread(struct thread_instance* lispthread)
+{
+#ifdef LISP_FEATURE_WIN32
+#define sb_pthr_kill(t,sig) \
+ __sync_fetch_and_or(&thread_extra_data(t)->pending_signal_set, 1<<sig)
+    /* META: why is this comment about safepoint builds mentioning
+     * gc_stop_the_world() ? Never the twain shall meet. */
+
+    /* Kludge (on safepoint builds): At the moment, this isn't just
+     * an optimization; rather it masks the fact that
+     * gc_stop_the_world() grabs the all_threads mutex without
+     * releasing it, and since we're not using recursive pthread
+     * mutexes, the pthread_mutex_lock() around the all_threads loop
+     * would go wrong.  Why are we running interruptions while
+     * stopping the world though?  Test case is (:ASYNC-UNWIND
+     * :SPECIALS), especially with s/10/100/ in both loops. */
+
+    /* Frequent special case: resignalling to self.  The idea is
+     * that leave_region safepoint will acknowledge the signal, so
+     * there is no need to take locks, roll thread to safepoint
+     * etc. */
+    struct thread* thread = (void*)lispthread->uw_primitive_thread;
+    if (thread == get_sb_vm_thread()) {
+        sb_pthr_kill(thread, 1); // can't fail
+        check_pending_thruptions(NULL);
+        return;
+    }
+    // block_deferrables + mutex_lock looks very unnecessary here,
+    // but without them, make-target-contrib hangs in bsd-sockets.
+    sigset_t oldset;
+    block_deferrable_signals(&oldset);
+    mutex_acquire(&all_threads_lock);
+    sb_pthr_kill(thread, 1); // can't fail
+    wake_thread_impl(lispthread);
+    mutex_release(&all_threads_lock);
+    thread_sigmask(SIG_SETMASK,&oldset,0);
+#else
+    wake_thread_impl(lispthread);
+#endif
+}
 
 void* os_get_csp(struct thread* th)
 {
@@ -987,9 +1016,8 @@ void* os_get_csp(struct thread* th)
 }
 
 
-#ifndef LISP_FEATURE_WIN32
+#ifdef LISP_FEATURE_UNIX
 
-# ifdef LISP_FEATURE_SB_SAFEPOINT
 /* This is basically what 'low_level_maybe_now_maybe_later' was (which doesn't exist),
  * but with a different name, and different way of deciding to defer the signal */
 void thruption_handler(__attribute__((unused)) int signal,
@@ -1014,7 +1042,6 @@ void thruption_handler(__attribute__((unused)) int signal,
     thread_in_lisp_raised(ctx);
     csp_around_foreign_call(self) = (intptr_t) transition_sp;
 }
-# endif
 
 #ifdef LISP_FEATURE_C_STACK_IS_CONTROL_STACK
 /* Trap trampolines are in target-assem.S so that they pick up the
@@ -1043,7 +1070,7 @@ handle_safepoint_violation(os_context_t *ctx, os_vm_address_t fault_address)
         return 1;
     }
 
-    if ((1+THREAD_HEADER_SLOTS)+(lispobj*)fault_address == (lispobj*)self) {
+    if (1+(lispobj*)fault_address == (lispobj*)self) {
 #ifdef LISP_FEATURE_C_STACK_IS_CONTROL_STACK
         arrange_return_to_c_function(ctx, handle_csp_safepoint_violation, 0);
 #else
@@ -1056,7 +1083,7 @@ handle_safepoint_violation(os_context_t *ctx, os_vm_address_t fault_address)
     /* not a safepoint */
     return 0;
 }
-#endif /* LISP_FEATURE_WIN32 */
+#endif /* LISP_FEATURE_UNIX */
 
 void
 vodxprint_fun(const char *fmt, va_list args)
